//#define DEBUGCONSOLE

using System.Collections.Generic;
using PAT.Common.Classes.DataStructure;
using PAT.Common.Classes.Expressions.ExpressionClass;
using PAT.Common.Classes.Ultility;
using PAT.PAR.Assertions;

namespace PAT.PAR.LTS
{
    public sealed class PARTimedInterrupt : PARProcess
    {
        public PARProcess FirstProcess;
        public PARProcess SecondProcess;
        public Variable Parameter;
        public byte TimerID;

        public PARTimedInterrupt(PARProcess firstProcess, PARProcess secondProcess, Variable parameter, byte timerID)
        {
            FirstProcess = firstProcess;
            SecondProcess = secondProcess;
            Parameter = parameter;
            TimerID = timerID;
            ProcessID = PARDataStore.DataManager.InitializeProcessID(FirstProcess.ProcessID + Constants.INTERRUPT + "[" + Parameter + "]" + SecondProcess.ProcessID);
        }

        public override List<PARConfiguration> MoveOneStep(PARConfiguration eStep)
        {
            // Get the parameter name
            string parameter_name = Parameter.ToString();
            // Add "clock <= param"
            eStep.AddParInequality(TimerID, TimerOperationType.LessThanOrEqualTo, parameter_name);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("PARTimedInterrupt:MoveOneStep > Adding PAR inequality 'x" + TimerID + " <= " + parameter_name + "'");
            System.Console.Out.WriteLine("PARTimedInterrupt:MoveOneStep > eStep " + eStep);
#endif

            List<PARConfiguration> steps = FirstProcess.MoveOneStep(eStep);

            foreach (PARConfiguration vm in steps)
            {
                vm.Process = new PARTimedInterrupt(vm.Process, SecondProcess, Parameter, TimerID);
            }

            Constraint newConstraint = eStep.TheConstraint.Clone();
            newConstraint.AddParInequality(TimerID, TimerOperationType.Equals, parameter_name);

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("PARTimedInterrupt:MoveOneStep > Adding PAR inequality 'x" + TimerID + " = " + parameter_name + "'");
            System.Console.Out.WriteLine("PARTimedInterrupt:MoveOneStep > " + newConstraint);
#endif
            steps.Add(new PARConfiguration(SecondProcess, Constants.TAU, "[interrupt[" + parameter_name + "]]", eStep.GlobalEnv, false, false, newConstraint));
            return steps;
        }

        public override List<PARConfigurationWithChannelData> SyncOutput(PARConfiguration eStep)
        {
            //get the int constant delay value
            //int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;
            List<PARConfigurationWithChannelData> steps = FirstProcess.SyncOutput(eStep);

            foreach (PARConfigurationWithChannelData vm in steps)
            {
                vm.Process = new PARTimedInterrupt(vm.Process, SecondProcess, Parameter, TimerID);
            }
            return steps;
        }

        public override List<PARConfiguration> SyncInput(PARConfigurationWithChannelData eStep)
        {
            //get the int constant delay value
            //int value = (EvaluatorDenotational.Evaluate(DelayTime, eStep.GlobalEnv) as IntConstant).Value;

            List<PARConfiguration> steps = FirstProcess.SyncInput(eStep);

            foreach (PARConfiguration vm in steps)
            {
                vm.Process = new PARTimedInterrupt(vm.Process, this.SecondProcess, Parameter, TimerID);
            }

            return steps;
        }

        public override List<string> GetGlobalVariables()
        {
            List<string> Variables = FirstProcess.GetGlobalVariables();
            Common.Classes.Ultility.Ultility.AddList<string>(Variables, SecondProcess.GetGlobalVariables());
            //Common.Classes.Ultility.Ultility.AddList(Variables, DelayTime.GetVars());
            return Variables;
        }

        public override List<string> GetChannels()
        {
            List<string> channel = FirstProcess.GetChannels();
            Common.Classes.Ultility.Ultility.AddList<string>(channel, SecondProcess.GetChannels());

            return channel;
        }

        public override string ToString()
        {
            if (TimerID == Constants.EMPTY_TIMER_ID)
            {
                return "(" + FirstProcess.ToString() + " interrupt[" + Parameter + "] " + SecondProcess.ToString() + ")";
            }
            else
            {
                return "(" + FirstProcess.ToString() + " interrupt[" + Parameter + "]" + Constraint.ClockToString(TimerID) + " " + SecondProcess.ToString() + ")";
            }
        }

        public override HashSet<string> GetAlphabets(Dictionary<string, string> visitedDefinitionRefs)
        {
            HashSet<string> list = SecondProcess.GetAlphabets(visitedDefinitionRefs);
            list.UnionWith(FirstProcess.GetAlphabets(visitedDefinitionRefs));
            return list;
        }

        public override PARProcess ClearConstant(Dictionary<string, Expression> constMapping)
        {
            return new PARTimedInterrupt(FirstProcess.ClearConstant(constMapping), SecondProcess.ClearConstant(constMapping), Parameter, TimerID);
        }

        public override bool MustBeAbstracted()
        {
            return FirstProcess.MustBeAbstracted() || SecondProcess.MustBeAbstracted();
        }

        //public override List<byte> GetActiveTimer()//Configuration eStep)
        //{
        //    List<byte> timers = new List<byte>();
        //    if (TimerID != Constants.EMPTY_TIMER_ID)
        //    {
        //        timers.Add(TimerID);
        //    }

        //    Common.Classes.Ultility.Ultility.Union(timers, FirstProcess.GetActiveTimer());//eStep));
        //    Common.Classes.Ultility.Ultility.Union(timers, SecondProcess.GetActiveTimer());//eStep));

        //    return timers;
        //}

        public override PARProcess SetTimer(byte timerID, ClockInfo clocks)
        {
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("Entering PARTimedInterrupt:SetTimer");
#endif
            PARProcess process = FirstProcess.SetTimer(timerID, clocks);

            // Get the parameter name
            string parameter_name = Parameter.ToString();

            if (process != null)
            {
                if (TimerID != Constants.EMPTY_TIMER_ID)
                {
                    clocks.AddParInequality(TimerID, parameter_name);
#if DEBUGCONSOLE
                    System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Addded to ClockInfo clocks a PAR inequality '" + Constraint.ClockToString(TimerID) + " <= " + parameter_name + "'");
                    System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > No activation, keep " + Constraint.ClockToString(TimerID));
#endif
            return new PARTimedInterrupt(process, SecondProcess, Parameter, TimerID);
                }

                clocks.AddParInequality(timerID, parameter_name);
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Addded to ClockInfo clocks a PAR inequality '" + Constraint.ClockToString(timerID) + " <= " + parameter_name + "'");
                System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Activation with " + Constraint.ClockToString(timerID));
#endif
                return new PARTimedInterrupt(process, SecondProcess, Parameter, timerID);
            }

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Process is null.");
#endif

            if (TimerID != Constants.EMPTY_TIMER_ID)
            {
                clocks.AddParInequality(TimerID, parameter_name);
#if DEBUGCONSOLE
                System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Addded to ClockInfo clocks a PAR inequality '" + Constraint.ClockToString(TimerID) + " <= " + parameter_name + "'");
                System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > return null, keep " + Constraint.ClockToString(TimerID));
#endif
                return null;
            }

            clocks.AddParInequality(timerID, parameter_name);
#if DEBUGCONSOLE
            System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Addded to ClockInfo clocks a PAR inequality '" + Constraint.ClockToString(timerID) + " <= " + parameter_name + "'");
            System.Console.Out.WriteLine("PARTimedInterrupt:SetTimer > Activation with " + Constraint.ClockToString(timerID));
#endif

#if DEBUGCONSOLE
            System.Console.Out.WriteLine("About to exit PARTimedInterrupt:SetTimer");
#endif

            return new PARTimedInterrupt(FirstProcess, SecondProcess, Parameter, timerID);
        }

        //public override bool IsTime()
        //{
        //    return true;
        //}
    }
}